%{
open Ast
%}

%token FUN LET REC COREC IN LOAD
%token EQ NEQ LT LE GT GE
%token PLUS MINUS MUL DIV CONCAT PLUSF MINUSF MULF DIVF MOD
%token COMMA
%token <int> INT
%token <float> FLOAT
%token <string> VAR
%token <string> CVAR
%token <string> STRING
%token IF THEN ELSE WHILE DO FOR TO DONE MATCH WITH
%token TRUE FALSE
%token SEQ IMP ASSG
%token CONS
%token LPAREN RPAREN LBRACKET RBRACKET
%token NOT AND OR
%token EOF
%token LOAD CHDIR PWD LS ENV SCOPE TYPE_INF QUIT
%token TYPEDEF OF BAR
%token <string> TYPE
%token TYPEDECL
%token <string> TYPEVAR

%nonassoc RPAREN RBRACKET /* lowest precedence - always reduce */
%right IMP
%left SEQ
%right ASSG
%left COMMA
%nonassoc CVAR
%left OR
%left AND
%left EQ NEQ LT LE GT GE
%right CONS
%left PLUS MINUS PLUSF MINUSF
%left MUL DIV MOD MULF DIVF CONCAT
%nonassoc NOT
%nonassoc INT FLOAT VAR TRUE FALSE STRING TYPE
%left BAR
%nonassoc LPAREN LBRACKET FUN LET IN ELSE IF WHILE FOR TYPEDECL TYPEVAR MATCH WITH
/* highest precedence - always shift */

/* entry points */
%start parse
%type <Ast.toplevel> parse

%%

parse:
  | directive EOF { Directive $1 }
  | declaration_list EOF { Declarations $1 }
  | expr EOF { Expression $1 }
;

directive:
  | LOAD STRING  { Load $2 }
  | CHDIR STRING { Chdir $2 }
  | PWD { Pwd }
  | LS  { Ls }
  | ENV { Env }
  | SCOPE { Scope }
  | TYPE_INF { Type_inf }
  | QUIT  { Quit }
;  

sumtype:
  | CVAR OF parse_type { ($1, Some($3)) }
  | CVAR { ($1, None) }
;

sumtypes:
  | sumtype BAR sumtypes { $1 :: $3 }
  | sumtype { [$1] }
;

declaration_list:
  | declaration declaration_list { $1 :: $2 }
  | declaration { [$1] }
;

declaration:
	| LET parse_var EQ expr %prec RPAREN
      { DefType ($2, Null, [], [], $4) }
	| LET parse_var TYPEDECL parse_type EQ expr %prec RPAREN
      { DefType ($2, $4, [], [], $6) }
	| LET LPAREN parse_var TYPEDECL parse_type RPAREN EQ expr %prec RPAREN
      { DefType ($3, $5, [], [], $8) }
	| LET parse_var typelist EQ expr %prec RPAREN
      { DefType($2, Null, fst (List.split $3), snd (List.split $3), $5) }
	| LET parse_var typelist TYPEDECL parse_type EQ expr %prec RPAREN
      { DefType($2, $5, fst (List.split $3), snd (List.split $3), $7) }
	| LET REC parse_var EQ expr %prec RPAREN
      { DefrecType ($3, Null, [], [], $5) }
	| LET REC parse_var TYPEDECL parse_type EQ expr %prec RPAREN
      { DefrecType ($3, $5, [], [], $7) }
	| LET REC LPAREN parse_var TYPEDECL parse_type RPAREN EQ expr %prec RPAREN
      { DefrecType ($4, $6, [], [], $9) }
	| LET REC parse_var typelist EQ expr %prec RPAREN
      { DefrecType($3, Null, fst (List.split $4), snd (List.split $4), $6) }
	| LET REC parse_var typelist TYPEDECL parse_type EQ expr %prec RPAREN
      { DefrecType($3, $6, fst (List.split $4), snd (List.split $4), $8) }
	| LET COREC LBRACKET expr RBRACKET parse_var EQ expr %prec RPAREN
      { DefcorecType ($4, $6, Null, [], [], $8) }
	| LET COREC LBRACKET expr RBRACKET parse_var TYPEDECL parse_type EQ expr %prec RPAREN
      { DefcorecType ($4, $6, $8, [], [], $10) }
	| LET COREC LBRACKET expr RBRACKET LPAREN parse_var TYPEDECL parse_type RPAREN EQ expr %prec RPAREN
      { DefcorecType ($4, $7, $9, [], [], $12) }
	| LET COREC LBRACKET expr RBRACKET parse_var typelist EQ expr %prec RPAREN
      { DefcorecType($4, $6, Null, fst (List.split $7), snd (List.split $7), $9) }
	| LET COREC LBRACKET expr RBRACKET parse_var typelist TYPEDECL parse_type EQ expr %prec RPAREN
      { DefcorecType($4, $6, $9, fst (List.split $7), snd (List.split $7), $11) }
        | TYPEDEF parse_var EQ sumtypes { Typedef($2, ([ ], $4)) }
        | TYPEDEF parse_var EQ BAR sumtypes { Typedef($2, ([ ], $5)) }
        | TYPEDEF typevar parse_var EQ sumtypes { Typedef($3, ($2, $5)) }
        | TYPEDEF typevar parse_var EQ BAR sumtypes { Typedef($3, ($2, $6)) }
;

typevar:
   | TYPEVAR { [ $1 ] }
   | LPAREN typevars RPAREN { $2 }
;

typevars: 
   | TYPEVAR { [ $1 ] }
   | TYPEVAR COMMA typevars { $1 :: $3 }
;

parse_var:
   | VAR { $1 }
   | LPAREN parse_var RPAREN { $2 }
;

param:
   | LPAREN VAR TYPEDECL parse_type RPAREN { (Var $2, $4) }
   | VAR { (Var $1, Null) }
   | LPAREN LPAREN RPAREN TYPEDECL parse_type RPAREN { (Unit, $5) }
   | LPAREN RPAREN { (Unit, Null) }
   | LPAREN param RPAREN { $2 }
;

paramlist:
   | param paramlist { $1 :: $2 }
   | param { [$1] }
;

func_and_params:
   | VAR TYPEDECL parse_type { [(Var $1, $3)] }
   | VAR { [(Var $1, Null)] }
   | VAR paramlist TYPEDECL parse_type { (Var $1, $4) :: $2 }
   | VAR paramlist { (Var $1, Null) :: $2 }
;

typelist:   
   | LPAREN parse_var TYPEDECL parse_type RPAREN { [(Var $2, $4)] }
   | LPAREN parse_var TYPEDECL parse_type RPAREN typelist { (Var $2, $4) :: $6}
   | parse_var { [(Var $1, Null)] } 
   | parse_var typelist { (Var $1, Null) :: $2 }
   | LPAREN LPAREN RPAREN TYPEDECL parse_type RPAREN { [(Unit, $5)] }
   | LPAREN LPAREN RPAREN TYPEDECL parse_type RPAREN typelist { (Unit, $5) :: $7 }
   | LPAREN RPAREN { [(Unit, Null)] }
   | LPAREN RPAREN typelist { (Unit, Null) :: $3 }
;

product_type :
  | product_type MUL parse_type { $3 :: $1 } /* shift/reduce conflict */
  | parse_type MUL parse_type { [ $3 ; $1 ] }
;

parse_types :
  | parse_type COMMA parse_types { $1 :: $3 }
  | parse_type { [ $1 ] }
;

typevar_list :
  | TYPEVAR { [ VarType $1 ] }
  | TYPEVAR COMMA typevar_list { (VarType $1) :: $3 }

parse_type : 
	| TYPE { match $1 with
	    | "int" -> Integer
	    | "string" -> String
	    | "bool" -> Boolean
	    | "unit" -> UnitType
            | "float" -> FloatType
	    | _ -> raise Parse_error }
        | VAR { UserType($1, [ ]) }
	| parse_type VAR { UserType($2, [ $1 ]) }
	| LPAREN parse_types RPAREN VAR { UserType($4, $2) }
	| TYPEVAR { VarType $1 }
	| TYPEVAR VAR { UserType($2, [ VarType $1 ]) }
	| LPAREN typevar_list RPAREN VAR { UserType ($4, $2) }
	| parse_type IMP parse_type { Arrow ($1, $3) }
	| LPAREN parse_type RPAREN { $2 }
        | product_type { TupleType (List.rev $1) }
;

arg:
  | VAR { Var $1 }
  | INT { Int $1 }
  | FLOAT { Float $1 }
  | STRING { Strg $1 }
  | TRUE { Bool true }
  | FALSE { Bool false }
  | LPAREN RPAREN { Unit }
  | LPAREN expr RPAREN { $2 }
  | LBRACKET RBRACKET { Inj( "[]", None ) }
  | LBRACKET expr RBRACKET {
    (* solving the ambiguity on the semicolumn, which can be in a list
       or a sequence of instructions *)
    let rec convert tail = function
    (* ; is left associative whereas :: is right associative:
       we need to reverse everything *)
      | Seq(hd, tl) ->
	convert (Inj("::", Some(Tuple [ tl; tail ]))) hd
      | e -> Inj("::", Some (Tuple [e; tail])) 
    in let res = convert (Inj("[]", None)) $2 in 
       res
  }
;

tuple:
  | tuple COMMA expr { $3 :: $1 } /* shift/reduce conflict */
  | expr COMMA expr { [ $3 ; $1 ] }
;

pattern_list:
  | pattern { PInj("::", Some(PTuple [$1; PInj("[]", None)])) }
  | pattern SEQ pattern_list { PInj("::", Some(PTuple [$1; $3])) }
;

pattern_tuple:
  | pattern_tuple COMMA pattern { $3 :: $1 }
  | pattern COMMA pattern { [ $3; $1 ] }
;

pattern:
  | LPAREN pattern RPAREN { $2 }
  | VAR { PVar $1 }
  | INT { PInt $1 }
  | FLOAT { PFloat $1 }
  | STRING { PStrg $1 }
  | TRUE { PBool true }
  | FALSE { PBool false }
  | LPAREN RPAREN { PUnit }
  | pattern CONS pattern { PInj("::", Some (PTuple [$1; $3])) }
  | LBRACKET RBRACKET { PInj("[]", None) }
  | LBRACKET pattern_list RBRACKET { $2 }
  | pattern_tuple { PTuple (List.rev $1) }
  | CVAR { PInj($1, None) }
  | CVAR pattern { PInj($1, Some($2)) }
;

pattern_expr:
  | pattern IMP expr { ($1, $3) }
;

patterns_expr:
  | pattern_expr { [ $1 ] }
  | pattern_expr BAR patterns_expr { $1 :: $3 }

patterns:
  | patterns_expr { $1 }
  | BAR patterns_expr { $2 }
;

expr:
  | FUN typelist IMP expr %prec RPAREN { FunType ( fst ( List.split $2), snd (List.split $2), $4) }
  | declaration IN expr %prec RPAREN {
    match $1 with
      | DefType (x, typeX, a, typeA, e) -> LetType (x, typeX, a, typeA, e, $3)
      | DefrecType (x, typeX, a, typeA, e) -> LetrecType (x, typeX, a, typeA, e, $3)
      | DefcorecType (s, x, typeX, a, typeA, e) -> LetcorecType(s, x, typeX, a, typeA, e, $3)
      | Def (x,a,e) -> raise Parse_error 
      | Defrec (x,a,e) -> raise Parse_error
      | Defcorec _ -> raise Parse_error
      | Typedef _ -> raise Parse_error
    }
  | VAR ASSG expr { Assg ($1, $3) }
  | expr SEQ expr { Seq ($1, $3) }
  | IF expr THEN expr ELSE expr %prec RPAREN { If ($2, $4, $6) }
  | WHILE expr DO expr DONE { While ($2, $4) }
  | FOR VAR EQ expr TO expr DO expr DONE { For ($2, $4, $6, $8) }
  | expr EQ expr { Eq ($1, $3) }
  | expr NEQ expr { Neq ($1, $3) }
  | expr LT expr { Lt ($1, $3) }
  | expr LE expr { Le ($1, $3) }
  | expr GT expr { Gt ($1, $3) }
  | expr GE expr { Ge ($1, $3) }
  | expr PLUS expr { Plus ($1, $3) }
  | expr MINUS expr { Minus ($1, $3) }
  | expr MUL expr { Mul ($1, $3) }
  | expr DIV expr { Div ($1, $3) }
  | expr MOD expr { Mod ($1, $3) }
  | expr PLUSF expr { PlusF ($1, $3) }
  | expr MINUSF expr { MinusF ($1, $3) }
  | expr MULF expr { MulF ($1, $3) }
  | expr DIVF expr { DivF ($1, $3) }
  | expr CONCAT expr { Concat ($1, $3) }
  | MINUS expr { Minus (Int 0, $2) }
  | MINUSF expr { MinusF (Float 0., $2) }
  | NOT expr { Not ($2) }
  | expr AND expr { And ($1, $3) }
  | expr OR expr { Or ($1, $3) }
  | expr CONS expr { Inj("::", Some (Tuple [$1; $3])) }
  | expr arg { App ($1, $2) }
  | arg { $1 }
  | tuple { Tuple(List.rev $1) }
  | CVAR { Inj($1, None) } /* introduces 8 shift/reduce conflicts */
  | CVAR arg { Inj($1, Some($2)) }
  | MATCH expr WITH patterns { Match($2, $4) }
;

%%
